@@ -76,88 +76,21 @@ display_control_sequence(const char *name, GValueArray *params)
 /* A couple are duplicated from vte.c, to keep them static... */
 
 /* Find the character an the given position in the backscroll buffer. */
-static struct vte_charcell *
-vte_terminal_find_charcell(VteTerminal *terminal, glong col, glong row)
+static VteCell *
+vte_terminal_find_charcell (VteTerminal *terminal, glong col, glong row)
 {
 	VteRowData *rowdata;
-	struct vte_charcell *ret = NULL;
+	VteCell *ret = NULL;
 	VteScreen *screen;
 	g_assert(VTE_IS_TERMINAL(terminal));
 	screen = terminal->pvt->screen;
-	if (_vte_ring_contains(screen->row_data, row)) {
-		rowdata = _vte_ring_index(screen->row_data, VteRowData *, row);
-		if ((glong) rowdata->cells->len > col) {
-			ret = &g_array_index(rowdata->cells,
-					     struct vte_charcell,
-					     col);
-		}
+	if (_vte_ring_contains (screen->row_data, row)) {
+		rowdata = _vte_ring_index_writable (screen->row_data, row);
+		ret = _vte_row_data_get_writable (rowdata, col);
 	}
 	return ret;
 }
 
-/* Append a single item to a GArray a given number of times. Centralizing all
- * of the places we do this may let me do something more clever later.
- * Dupped from vte.c. */
-static void
-vte_g_array_fill(GArray *array, gpointer item, guint final_size)
-{
-	g_assert(array != NULL);
-	if (array->len >= final_size) {
-		return;
-	}
-	g_assert(item != NULL);
-
-	final_size -= array->len;
-	do {
-		g_array_append_vals(array, item, 1);
-	} while (--final_size);
-}
-
-/* Insert a blank line at an arbitrary position. */
-static void
-vte_insert_line_internal(VteTerminal *terminal, glong position)
-{
-	VteRowData *row, *old_row;
-	old_row = terminal->pvt->free_row;
-	/* Pad out the line data to the insertion point. */
-	while (_vte_ring_next(terminal->pvt->screen->row_data) < position) {
-		if (old_row) {
-			row = _vte_reset_row_data (terminal, old_row, TRUE);
-		} else {
-			row = _vte_new_row_data_sized(terminal, TRUE);
-		}
-		old_row = _vte_ring_append(terminal->pvt->screen->row_data, row);
-	}
-	/* If we haven't inserted a line yet, insert a new one. */
-	if (old_row) {
-		row = _vte_reset_row_data (terminal, old_row, TRUE);
-	} else {
-		row = _vte_new_row_data_sized(terminal, TRUE);
-	}
-	if (_vte_ring_next(terminal->pvt->screen->row_data) >= position) {
-		old_row = _vte_ring_insert(terminal->pvt->screen->row_data,
-				 position, row);
-	} else {
-		old_row =_vte_ring_append(terminal->pvt->screen->row_data, row);
-	}
-	terminal->pvt->free_row = old_row;
-}
-
-/* Remove a line at an arbitrary position. */
-static void
-vte_remove_line_internal(VteTerminal *terminal, glong position)
-{
-	if (_vte_ring_next(terminal->pvt->screen->row_data) > position) {
-		if (terminal->pvt->free_row)
-			_vte_free_row_data (terminal->pvt->free_row);
-
-		terminal->pvt->free_row = _vte_ring_remove(
-				terminal->pvt->screen->row_data,
-				position,
-				FALSE);
-	}
-}
-
 /* Check how long a string of unichars is.  Slow version. */
 static gssize
 vte_unichar_strlen(gunichar *c)
@@ -183,7 +116,7 @@ vte_ucs4_to_utf8 (VteTerminal *terminal, const guchar *in)
 		outlen = (inlen * VTE_UTF8_BPC) + 1;
 
 		_vte_buffer_set_minimum_size (terminal->pvt->conv_buffer, outlen);
-		buf = bufptr = terminal->pvt->conv_buffer->bytes;
+		buf = bufptr = terminal->pvt->conv_buffer->data;
 
 		if (_vte_conv (conv, &in, &inlen, &buf, &outlen) == (size_t) -1) {
 			_vte_debug_print (VTE_DEBUG_IO,
@@ -336,27 +269,15 @@ _vte_terminal_home_cursor (VteTerminal *terminal)
 static void
 _vte_terminal_clear_screen (VteTerminal *terminal)
 {
-	VteRowData *rowdata, *old_row;
 	long i, initial, row;
 	VteScreen *screen;
 	screen = terminal->pvt->screen;
 	initial = screen->insert_delta;
 	row = screen->cursor_current.row - screen->insert_delta;
+	initial = _vte_ring_next(screen->row_data);
 	/* Add a new screen's worth of rows. */
-	old_row = terminal->pvt->free_row;
-	for (i = 0; i < terminal->row_count; i++) {
-		/* Add a new row */
-		if (i == 0) {
-			initial = _vte_ring_next(screen->row_data);
-		}
-		if (old_row) {
-			rowdata = _vte_reset_row_data (terminal, old_row, TRUE);
-		} else {
-			rowdata = _vte_new_row_data_sized(terminal, TRUE);
-		}
-		old_row = _vte_ring_append(screen->row_data, rowdata);
-	}
-	terminal->pvt->free_row = old_row;
+	for (i = 0; i < terminal->row_count; i++)
+		_vte_ring_append(screen->row_data);
 	/* Move the cursor and insertion delta to the first line in the
 	 * newly-cleared area and scroll if need be. */
 	screen->insert_delta = initial;
@@ -381,18 +302,12 @@ _vte_terminal_clear_current_line (VteTerminal *terminal)
 	 * which corresponds to the cursor. */
 	if (_vte_ring_next(screen->row_data) > screen->cursor_current.row) {
 		/* Get the data for the row which the cursor points to. */
-		rowdata = _vte_ring_index(screen->row_data, VteRowData *,
-					  screen->cursor_current.row);
+		rowdata = _vte_ring_index_writable (screen->row_data, screen->cursor_current.row);
 		g_assert(rowdata != NULL);
 		/* Remove it. */
-		if (rowdata->cells->len > 0) {
-			g_array_set_size(rowdata->cells, 0);
-		}
-		/* Add enough cells to the end of the line to fill out the
-		 * row. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_row_data_shrink (rowdata, 0);
+		/* Add enough cells to the end of the line to fill out the row. */
+		_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 		rowdata->soft_wrapped = 0;
 		/* Repaint this row. */
 		_vte_invalidate_cells(terminal,
@@ -416,20 +331,13 @@ _vte_terminal_clear_above_current (VteTerminal *terminal)
 	 * which corresponds to the cursor. */
 	for (i = screen->insert_delta; i < screen->cursor_current.row; i++) {
 		if (_vte_ring_next(screen->row_data) > i) {
-			guint len;
 			/* Get the data for the row we're erasing. */
-			rowdata = _vte_ring_index(screen->row_data,
-						  VteRowData *, i);
+			rowdata = _vte_ring_index_writable (screen->row_data, i);
 			g_assert(rowdata != NULL);
 			/* Remove it. */
-			len = rowdata->cells->len;
-			if (len > 0) {
-				g_array_set_size(rowdata->cells, 0);
-			}
+			_vte_row_data_shrink (rowdata, 0);
 			/* Add new cells until we fill the row. */
-			vte_g_array_fill(rowdata->cells,
-					 &screen->fill_defaults,
-					 terminal->column_count);
+			_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 			rowdata->soft_wrapped = 0;
 			/* Repaint the row. */
 			_vte_invalidate_cells(terminal,
@@ -444,7 +352,6 @@ _vte_terminal_clear_above_current (VteTerminal *terminal)
 static void
 _vte_terminal_scroll_text (VteTerminal *terminal, int scroll_amount)
 {
-	VteRowData *row, *old_row;
 	long start, end, i;
 	VteScreen *screen;
 
@@ -458,25 +365,18 @@ _vte_terminal_scroll_text (VteTerminal *terminal, int scroll_amount)
 		end = start + terminal->row_count - 1;
 	}
 
-	old_row = terminal->pvt->free_row;
-	while (_vte_ring_next(screen->row_data) <= end) {
-		if (old_row) {
-			row = _vte_reset_row_data (terminal, old_row, FALSE);
-		} else {
-			row = _vte_new_row_data_sized(terminal, FALSE);
-		}
-		old_row = _vte_ring_append(terminal->pvt->screen->row_data, row);
-	}
-	terminal->pvt->free_row = old_row;
+	while (_vte_ring_next(screen->row_data) <= end)
+		_vte_ring_append(terminal->pvt->screen->row_data);
+
 	if (scroll_amount > 0) {
 		for (i = 0; i < scroll_amount; i++) {
-			vte_remove_line_internal(terminal, end);
-			vte_insert_line_internal(terminal, start);
+			_vte_ring_remove (terminal->pvt->screen->row_data, end);
+			_vte_ring_insert (terminal->pvt->screen->row_data, start);
 		}
 	} else {
 		for (i = 0; i < -scroll_amount; i++) {
-			vte_remove_line_internal(terminal, start);
-			vte_insert_line_internal(terminal, end);
+			_vte_ring_remove (terminal->pvt->screen->row_data, start);
+			_vte_ring_insert (terminal->pvt->screen->row_data, end);
 		}
 	}
 
@@ -895,6 +795,7 @@ vte_sequence_handler_decset_internal(VteTerminal *terminal,
 				"Entering application cursor mode.\n" :
 				"Leaving application cursor mode.\n");
 		break;
+#if 0		/* 3: disallowed, window size is set by user. */
 	case 3:
 		vte_terminal_emit_resize_window(terminal,
 						(set ? 132 : 80) *
@@ -906,6 +807,7 @@ vte_sequence_handler_decset_internal(VteTerminal *terminal,
 		/* Request a resize and redraw. */
 		_vte_invalidate_all(terminal);
 		break;
+#endif
 	case 5:
 		/* Repaint everything in reverse mode. */
 		_vte_invalidate_all(terminal);
@@ -1021,15 +923,10 @@ vte_sequence_handler_al (VteTerminal *terminal, GValueArray *params)
 	for (i = 0; i < param; i++) {
 		/* Clear a line off the end of the region and add one to the
 		 * top of the region. */
-		vte_remove_line_internal(terminal, end);
-		vte_insert_line_internal(terminal, start);
-		/* Get the data for the new row. */
-		rowdata = _vte_ring_index(screen->row_data,
-					  VteRowData *, start);
-		g_assert(rowdata != NULL);
+		_vte_ring_remove (terminal->pvt->screen->row_data, end);
+		rowdata = _vte_ring_insert (terminal->pvt->screen->row_data, start);
 		/* Add enough cells to it so that it has the default columns. */
-		vte_g_array_fill(rowdata->cells, &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 		/* Adjust the scrollbars if necessary. */
 		_vte_terminal_adjust_adjustments(terminal);
 	}
@@ -1101,7 +998,7 @@ vte_sequence_handler_cb (VteTerminal *terminal, GValueArray *params)
 	VteRowData *rowdata;
 	long i;
 	VteScreen *screen;
-	struct vte_charcell *pcell;
+	VteCell *pcell;
 	screen = terminal->pvt->screen;
 
 	/* Get the data for the row which the cursor points to. */
@@ -1110,16 +1007,13 @@ vte_sequence_handler_cb (VteTerminal *terminal, GValueArray *params)
 	 * attributes.  If there is no such character cell, we need
 	 * to add one. */
 	for (i = 0; i <= screen->cursor_current.col; i++) {
-		if (i < (glong) rowdata->cells->len) {
+		if (i < (glong) _vte_row_data_length (rowdata)) {
 			/* Muck with the cell in this location. */
-			pcell = &g_array_index(rowdata->cells,
-					       struct vte_charcell,
-					       i);
+			pcell = _vte_row_data_get_writable (rowdata, i);
 			*pcell = screen->color_defaults;
 		} else {
 			/* Add new cells until we have one here. */
-			g_array_append_val(rowdata->cells,
-					   screen->color_defaults);
+			_vte_row_data_append (rowdata, &screen->color_defaults);
 		}
 	}
 	/* Repaint this row. */
@@ -1145,49 +1039,34 @@ vte_sequence_handler_cd (VteTerminal *terminal, GValueArray *params)
 	i = screen->cursor_current.row;
 	if (i < _vte_ring_next(screen->row_data)) {
 		/* Get the data for the row we're clipping. */
-		rowdata = _vte_ring_index(screen->row_data, VteRowData *, i);
+		rowdata = _vte_ring_index_writable (screen->row_data, i);
 		/* Clear everything to the right of the cursor. */
-		if ((rowdata != NULL) &&
-		    ((glong) rowdata->cells->len > screen->cursor_current.col)) {
-			g_array_set_size(rowdata->cells,
-					 screen->cursor_current.col);
-		}
+		if (rowdata)
+			_vte_row_data_shrink (rowdata, screen->cursor_current.col);
 	}
 	/* Now for the rest of the lines. */
 	for (i = screen->cursor_current.row + 1;
 	     i < _vte_ring_next(screen->row_data);
 	     i++) {
 		/* Get the data for the row we're removing. */
-		rowdata = _vte_ring_index(screen->row_data, VteRowData *, i);
+		rowdata = _vte_ring_index_writable (screen->row_data, i);
 		/* Remove it. */
-		if ((rowdata != NULL) && (rowdata->cells->len > 0)) {
-			g_array_set_size(rowdata->cells, 0);
-		}
+		if (rowdata)
+			_vte_row_data_shrink (rowdata, 0);
 	}
 	/* Now fill the cleared areas. */
 	for (i = screen->cursor_current.row;
 	     i < screen->insert_delta + terminal->row_count;
 	     i++) {
 		/* Retrieve the row's data, creating it if necessary. */
-		if (_vte_ring_contains(screen->row_data, i)) {
-			rowdata = _vte_ring_index(screen->row_data,
-						  VteRowData *, i);
+		if (_vte_ring_contains (screen->row_data, i)) {
+			rowdata = _vte_ring_index_writable (screen->row_data, i);
 			g_assert(rowdata != NULL);
 		} else {
-			if (terminal->pvt->free_row) {
-				rowdata = _vte_reset_row_data (terminal,
-						terminal->pvt->free_row,
-						FALSE);
-			} else {
-				rowdata = _vte_new_row_data(terminal);
-			}
-			terminal->pvt->free_row =
-				_vte_ring_append(screen->row_data, rowdata);
+			rowdata = _vte_ring_append (screen->row_data);
 		}
 		/* Pad out the row. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 		rowdata->soft_wrapped = 0;
 		/* Repaint this row. */
 		_vte_invalidate_cells(terminal,
@@ -1212,16 +1091,14 @@ vte_sequence_handler_ce (VteTerminal *terminal, GValueArray *params)
 	g_assert(rowdata != NULL);
 	/* Remove the data at the end of the array until the current column
 	 * is the end of the array. */
-	if ((glong) rowdata->cells->len > screen->cursor_current.col) {
-		g_array_set_size(rowdata->cells, screen->cursor_current.col);
+	if ((glong) _vte_row_data_length (rowdata) > screen->cursor_current.col) {
+		_vte_row_data_shrink (rowdata, screen->cursor_current.col);
 		/* We've modified the display.  Make a note of it. */
 		terminal->pvt->text_deleted_flag = TRUE;
 	}
 	if (screen->fill_defaults.attr.back != VTE_DEF_BG) {
 		/* Add enough cells to fill out the row. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 	}
 	rowdata->soft_wrapped = 0;
 	/* Repaint this row. */
@@ -1473,19 +1350,15 @@ vte_sequence_handler_dc (VteTerminal *terminal, GValueArray *params)
 	if (_vte_ring_next(screen->row_data) > screen->cursor_current.row) {
 		long len;
 		/* Get the data for the row which the cursor points to. */
-		rowdata = _vte_ring_index(screen->row_data,
-					  VteRowData *,
-					  screen->cursor_current.row);
+		rowdata = _vte_ring_index_writable (screen->row_data, screen->cursor_current.row);
 		g_assert(rowdata != NULL);
 		col = screen->cursor_current.col;
-		len = rowdata->cells->len;
+		len = _vte_row_data_length (rowdata);
 		/* Remove the column. */
 		if (col < len) {
-			g_array_remove_index(rowdata->cells, col);
+			_vte_row_data_remove (rowdata, col);
 			if (screen->fill_defaults.attr.back != VTE_DEF_BG) {
-				vte_g_array_fill (rowdata->cells,
-						&screen->fill_defaults,
-						terminal->column_count);
+				_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 				len = terminal->column_count;
 			}
 			/* Repaint this row. */
@@ -1536,8 +1409,8 @@ vte_sequence_handler_dl (VteTerminal *terminal, GValueArray *params)
 	for (i = 0; i < param; i++) {
 		/* Clear a line off the end of the region and add one to the
 		 * top of the region. */
-		vte_remove_line_internal(terminal, start);
-		vte_insert_line_internal(terminal, end);
+		_vte_ring_remove (terminal->pvt->screen->row_data, start);
+		_vte_ring_insert (terminal->pvt->screen->row_data, end);
 		/* Adjust the scrollbars if necessary. */
 		_vte_terminal_adjust_adjustments(terminal);
 	}
@@ -1599,7 +1472,7 @@ vte_sequence_handler_ec (VteTerminal *terminal, GValueArray *params)
 	VteScreen *screen;
 	VteRowData *rowdata;
 	GValue *value;
-	struct vte_charcell *cell;
+	VteCell *cell;
 	long col, i, count;
 
 	screen = terminal->pvt->screen;
@@ -1622,18 +1495,14 @@ vte_sequence_handler_ec (VteTerminal *terminal, GValueArray *params)
 		for (i = 0; i < count; i++) {
 			col = screen->cursor_current.col + i;
 			if (col >= 0) {
-				if (col < (glong) rowdata->cells->len) {
+				if (col < (glong) _vte_row_data_length (rowdata)) {
 					/* Replace this cell with the current
 					 * defaults. */
-					cell = &g_array_index(rowdata->cells,
-							      struct vte_charcell,
-							      col);
+					cell = _vte_row_data_get_writable (rowdata, col);
 					*cell = screen->color_defaults;
 				} else {
 					/* Add new cells until we have one here. */
-					vte_g_array_fill(rowdata->cells,
-							 &screen->color_defaults,
-							 col);
+					_vte_row_data_fill (rowdata, &screen->color_defaults, col);
 				}
 			}
 		}
@@ -2109,8 +1978,8 @@ vte_sequence_handler_sr (VteTerminal *terminal, GValueArray *params)
 	if (screen->cursor_current.row == start) {
 		/* If we're at the top of the scrolling region, add a
 		 * line at the top to scroll the bottom off. */
-		vte_remove_line_internal(terminal, end);
-		vte_insert_line_internal(terminal, start);
+		_vte_ring_remove (terminal->pvt->screen->row_data, end);
+		_vte_ring_insert (terminal->pvt->screen->row_data, start);
 		/* Update the display. */
 		_vte_terminal_scroll_region(terminal, start, end - start + 1, 1);
 		_vte_invalidate_cells(terminal,
@@ -2188,26 +2057,23 @@ vte_sequence_handler_ta (VteTerminal *terminal, GValueArray *params)
 		 */
 
 		/* Get rid of trailing empty cells: bug 545924 */
-		if ((glong) rowdata->cells->len > col)
+		if ((glong) _vte_row_data_length (rowdata) > col)
 		{
-			struct vte_charcell *cell;
+			const VteCell *cell;
 			guint i;
-			for (i = rowdata->cells->len; (glong) i > col; i--) {
-				cell = &g_array_index(rowdata->cells,
-						      struct vte_charcell, i - 1);
+			for (i = _vte_row_data_length (rowdata); (glong) i > col; i--) {
+				cell = _vte_row_data_get (rowdata, i - 1);
 				if (cell->attr.fragment || cell->c != 0)
 					break;
 			}
-			g_array_set_size(rowdata->cells, i);
+			_vte_row_data_shrink (rowdata, i);
 		}
 
-		if ((glong) rowdata->cells->len <= col)
+		if ((glong) _vte_row_data_length (rowdata) <= col)
 		  {
-		    struct vte_charcell cell;
+		    VteCell cell;
 
-		    vte_g_array_fill (rowdata->cells,
-				      &screen->fill_defaults,
-				      col);
+		    _vte_row_data_fill (rowdata, &screen->fill_defaults, col);
 
 		    cell.attr = screen->fill_defaults.attr;
 		    cell.attr.invisible = 1; /* FIXME: bug 499944 */
@@ -2220,20 +2086,16 @@ vte_sequence_handler_ta (VteTerminal *terminal, GValueArray *params)
 		       goto fallback_tab;
 		      }
 		    cell.c = '\t';
-		    g_array_append_vals(rowdata->cells, &cell, 1);
+		    _vte_row_data_append (rowdata, &cell);
 
 		    cell.attr = screen->fill_defaults.attr;
 		    cell.attr.fragment = 1;
-		    vte_g_array_fill (rowdata->cells,
-				      &cell,
-				newcol);
+		    _vte_row_data_fill (rowdata, &cell, newcol);
 		  }
 		else
 		  {
 		  fallback_tab:
-		    vte_g_array_fill (rowdata->cells,
-				      &screen->fill_defaults,
-				      newcol);
+		    _vte_row_data_fill (rowdata, &screen->fill_defaults, newcol);
 		  }
 
 		_vte_invalidate_cells (terminal,
@@ -2282,20 +2144,16 @@ vte_sequence_handler_ts (VteTerminal *terminal, GValueArray *params)
 static void
 vte_sequence_handler_uc (VteTerminal *terminal, GValueArray *params)
 {
-	struct vte_charcell *cell;
+	VteCell *cell;
 	int column;
 	VteScreen *screen;
 
 	screen = terminal->pvt->screen;
 	column = screen->cursor_current.col;
-	cell = vte_terminal_find_charcell(terminal,
-					  column,
-					  screen->cursor_current.row);
+	cell = vte_terminal_find_charcell(terminal, column, screen->cursor_current.row);
 	while ((cell != NULL) && (cell->attr.fragment) && (column > 0)) {
 		column--;
-		cell = vte_terminal_find_charcell(terminal,
-						  column,
-						  screen->cursor_current.row);
+		cell = vte_terminal_find_charcell(terminal, column, screen->cursor_current.row);
 	}
 	if (cell != NULL) {
 		/* Set this character to be underlined. */
@@ -2902,15 +2760,10 @@ vte_sequence_handler_insert_lines (VteTerminal *terminal, GValueArray *params)
 	for (i = 0; i < param; i++) {
 		/* Clear a line off the end of the region and add one to the
 		 * top of the region. */
-		vte_remove_line_internal(terminal, end);
-		vte_insert_line_internal(terminal, row);
-		/* Get the data for the new row. */
-		rowdata = _vte_ring_index(screen->row_data, VteRowData *, row);
-		g_assert(rowdata != NULL);
+		_vte_ring_remove (terminal->pvt->screen->row_data, end);
+		rowdata = _vte_ring_insert (terminal->pvt->screen->row_data, row);
 		/* Add enough cells to it so that it has the default colors. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 	}
 	/* Update the display. */
 	_vte_terminal_scroll_region(terminal, row, end - row + 1, param);
@@ -2951,15 +2804,10 @@ vte_sequence_handler_delete_lines (VteTerminal *terminal, GValueArray *params)
 	for (i = 0; i < param; i++) {
 		/* Insert a line at the end of the region and remove one from
 		 * the top of the region. */
-		vte_remove_line_internal(terminal, row);
-		vte_insert_line_internal(terminal, end);
-		/* Get the data for the new row. */
-		rowdata = _vte_ring_index(screen->row_data, VteRowData *, end);
-		g_assert(rowdata != NULL);
+		_vte_ring_remove (terminal->pvt->screen->row_data, row);
+		rowdata = _vte_ring_insert (terminal->pvt->screen->row_data, end);
 		/* Add enough cells to it so that it has the default colors. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 	}
 	/* Update the display. */
 	_vte_terminal_scroll_region(terminal, row, end - row + 1, -param);
@@ -3114,9 +2962,9 @@ static void
 vte_sequence_handler_screen_alignment_test (VteTerminal *terminal, GValueArray *params)
 {
 	long row;
-	VteRowData *rowdata, *old_row;
+	VteRowData *rowdata;
 	VteScreen *screen;
-	struct vte_charcell cell;
+	VteCell cell;
 
 	screen = terminal->pvt->screen;
 
@@ -3124,29 +2972,20 @@ vte_sequence_handler_screen_alignment_test (VteTerminal *terminal, GValueArray *
 	     row < terminal->pvt->screen->insert_delta + terminal->row_count;
 	     row++) {
 		/* Find this row. */
-		old_row = terminal->pvt->free_row;
-		while (_vte_ring_next(screen->row_data) <= row) {
-			if (old_row) {
-				rowdata = _vte_reset_row_data (terminal, old_row, FALSE);
-			} else {
-				rowdata = _vte_new_row_data(terminal);
-			}
-			old_row = _vte_ring_append(screen->row_data, rowdata);
-		}
-		terminal->pvt->free_row = old_row;
+		while (_vte_ring_next(screen->row_data) <= row)
+			_vte_ring_append(screen->row_data);
 		_vte_terminal_adjust_adjustments(terminal);
-		rowdata = _vte_ring_index(screen->row_data, VteRowData *, row);
+		rowdata = _vte_ring_index_writable (screen->row_data, row);
 		g_assert(rowdata != NULL);
 		/* Clear this row. */
-		if (rowdata->cells->len > 0) {
-			g_array_set_size(rowdata->cells, 0);
-		}
+		_vte_row_data_shrink (rowdata, 0);
+
 		_vte_terminal_emit_text_deleted(terminal);
 		/* Fill this row. */
 		cell.c = 'E';
-		cell.attr = screen->basic_defaults.attr;
+		cell.attr = basic_cell.cell.attr;
 		cell.attr.columns = 1;
-		vte_g_array_fill(rowdata->cells, &cell, terminal->column_count);
+		_vte_row_data_fill (rowdata, &cell, terminal->column_count);
 		_vte_terminal_emit_text_inserted(terminal);
 	}
 	_vte_invalidate_all(terminal);
